/*
 * Copyright (C) 2013 4th Line GmbH, Switzerland
 *
 * The contents of this file are subject to the terms of either the GNU
 * Lesser General Public License Version 2 or later ("LGPL") or the
 * Common Development and Distribution License Version 1 or later
 * ("CDDL") (collectively, the "License"). You may not use this file
 * except in compliance with the License. See LICENSE.txt for more
 * information.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

package com.gctronic.upnp_igd;

import android.app.Activity;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.graphics.Color;
import android.os.Bundle;
import android.os.IBinder;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.widget.ImageView;
import android.widget.TextView;
import android.widget.Toast;

import org.fourthline.cling.android.AndroidUpnpService;
import org.fourthline.cling.android.AndroidUpnpServiceImpl;
import org.fourthline.cling.android.FixedAndroidLogHandler;
import org.fourthline.cling.model.action.ActionInvocation;
import org.fourthline.cling.model.message.UpnpResponse;
import org.fourthline.cling.model.meta.Device;
import org.fourthline.cling.model.meta.LocalDevice;
import org.fourthline.cling.model.meta.RemoteDevice;
import org.fourthline.cling.model.meta.Service;
import org.fourthline.cling.model.types.DeviceType;
import org.fourthline.cling.model.types.UDADeviceType;
import org.fourthline.cling.model.types.UDAServiceId;
import org.fourthline.cling.model.types.UDAServiceType;
import org.fourthline.cling.model.types.UDN;
import org.fourthline.cling.registry.DefaultRegistryListener;
import org.fourthline.cling.registry.Registry;
import org.fourthline.cling.support.igd.PortMappingListener;
import org.fourthline.cling.support.igd.callback.GetExternalIP;
import org.fourthline.cling.support.model.PortMapping;
import org.fourthline.cling.transport.Router;
import org.fourthline.cling.transport.RouterException;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Enumeration;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Christian Bauer
 */
// DOC:CLASS
public class LightActivity extends Activity implements PropertyChangeListener {

    private static final String TAG = LightActivity.class.getName();

    // DOC:CLASS
    private static final Logger log = Logger.getLogger(LightActivity.class.getName());

    // DOC:SERVICE_BINDING
    private AndroidUpnpService upnpService;

    private UDN udn = new UDN(UUID.randomUUID()); // TODO: Not stable!

    private BrowseRegistryListener registryListener = new BrowseRegistryListener();

    private TextView txtLocal, txtRemote;

    String myIp = getIpAddress();
    int port = 8080;
    Service routerService = null;

    //creates a port mapping configuration with the external/internal port, an internal host IP, the protocol and an optional description
    PortMapping desiredMapping = new PortMapping(port, myIp, PortMapping.Protocol.TCP, "Spydroid");

    private ServiceConnection serviceConnection = new ServiceConnection() {

        public void onServiceConnected(ComponentName className, IBinder service) {
            upnpService = (AndroidUpnpService) service;

            // Get ready for future device advertisements
            upnpService.getRegistry().addListener(registryListener);
            upnpService.getRegistry().addListener(new PortMappingListener(desiredMapping));

            // Now add all devices to the list we already know about
            for (Device device : upnpService.getRegistry().getDevices()) {
                registryListener.deviceAdded(device);
            }

            // Search asynchronously for all devices, they will respond soon
            upnpService.getControlPoint().search();
        }

        public void onServiceDisconnected(ComponentName className) {
            upnpService = null;
        }
    };

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        // DOC:LOGGING
        // Fix the logging integration between java.util.logging and Android internal logging
         org.seamless.util.logging.LoggingUtil.resetRootHandler(
            new FixedAndroidLogHandler()
         );
         Logger.getLogger("org.fourthline.cling").setLevel(Level.FINEST);
        // DOC:LOGGING

        setContentView(R.layout.light);

        txtLocal = (TextView)findViewById(R.id.txtLocal);
        txtRemote = (TextView)findViewById(R.id.txtRemote);

        txtLocal.setText("Local: " + myIp + ":" + port);

        getApplicationContext().bindService(
                new Intent(this, AndroidUpnpServiceImpl.class),
                serviceConnection,
                Context.BIND_AUTO_CREATE
        );
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        getApplicationContext().unbindService(serviceConnection);
    }

    // DOC:SERVICE_BINDING

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        menu.add(0, 0, 0, R.string.switchRouter).setIcon(android.R.drawable.ic_menu_revert);
        menu.add(0, 1, 0, R.string.toggleDebugLogging).setIcon(android.R.drawable.ic_menu_info_details);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case 0:
                if (upnpService != null) {
                    Router router = upnpService.get().getRouter();
                    try {
                        if (router.isEnabled()) {
                            Toast.makeText(this, R.string.disablingRouter, Toast.LENGTH_SHORT).show();
                            router.disable();
                        } else {
                            Toast.makeText(this, R.string.enablingRouter, Toast.LENGTH_SHORT).show();
                            router.enable();
                        }
                    } catch (RouterException ex) {
                        Toast.makeText(this, getText(R.string.errorSwitchingRouter) + ex.toString(), Toast.LENGTH_LONG).show();
                        ex.printStackTrace(System.err);
                    }
                }
                break;
            case 1:
                Logger logger = Logger.getLogger("org.fourthline.cling");
                if (logger.getLevel() != null && !logger.getLevel().equals(Level.INFO)) {
                    Toast.makeText(this, R.string.disablingDebugLogging, Toast.LENGTH_SHORT).show();
                    logger.setLevel(Level.INFO);
                } else {
                    Toast.makeText(this, R.string.enablingDebugLogging, Toast.LENGTH_SHORT).show();
                    logger.setLevel(Level.FINEST);
                }
                break;
        }
        return false;
    }

    // DOC:PROPERTY_CHANGE
    public void propertyChange(PropertyChangeEvent event) {
        // This is regular JavaBean eventing, not UPnP eventing!
        if (event.getPropertyName().equals("status")) {
            log.info("Turning light: " + event.getNewValue());
        }
    }


    protected class BrowseRegistryListener extends DefaultRegistryListener {

        /* Discovery performance optimization for very slow Android devices! */
        @Override
        public void remoteDeviceDiscoveryStarted(Registry registry, RemoteDevice device) {
            deviceAdded(device);
        }

        @Override
        public void remoteDeviceDiscoveryFailed(Registry registry, final RemoteDevice device, final Exception ex) {
            runOnUiThread(new Runnable() {
                public void run() {
                    Toast.makeText(
                            LightActivity.this,
                            "Discovery failed of '" + device.getDisplayString() + "': "
                                    + (ex != null ? ex.toString() : "Couldn't retrieve device/service descriptors"),
                            Toast.LENGTH_LONG
                    ).show();
                }
            });
            deviceRemoved(device);
        }
        /* End of optimization, you can remove the whole block if your Android handset is fast (>= 600 Mhz) */

        @Override
        public void remoteDeviceAdded(Registry registry, RemoteDevice device) {
            deviceAdded(device);
        }

        @Override
        public void remoteDeviceRemoved(Registry registry, RemoteDevice device) {
            deviceRemoved(device);
        }

        @Override
        public void localDeviceAdded(Registry registry, LocalDevice device) {
            deviceAdded(device);
        }

        @Override
        public void localDeviceRemoved(Registry registry, LocalDevice device) {
            deviceRemoved(device);
        }

        public void deviceAdded(final Device device) {
            Log.i(TAG, "DEVICE ADDED");
            Log.i(TAG, "identity: " + device.getIdentity().toString());
            Log.i(TAG, "details: " + device.getDetails().toString());
            Log.i(TAG, "services: " + device.getServices().toString());
            Service[] ser = device.getServices();
            for(int i=0; i<ser.length; i++) {
                Log.i(TAG, i + ": " + ser[i].toString() + ", id=" + ser[i].getServiceId().toString() + ", type=" + ser[i].getServiceType().toString() + ", typeFriendly=" + ser[i].getServiceType().toFriendlyString());
            }
            //device.findService()
            Log.i(TAG, "type: " + device.getType().toString());
            Log.i(TAG, "type details: type=" + device.getType().getType().toString() + ", ns=" + device.getType().getNamespace().toString() + ", vers=" + device.getType().getVersion());

            if(routerService==null && device.getType().equals(new UDADeviceType("InternetGatewayDevice", 1)) && device.isFullyHydrated()) {
                Log.i(TAG, "Found router!");

                DeviceType CONNECTION_DEVICE_TYPE = new UDADeviceType("WANConnectionDevice", 1);
                Device[] connectionDevices = device.findDevices(CONNECTION_DEVICE_TYPE);
                if (connectionDevices.length == 0) {
                    log.fine("IGD doesn't support '" + CONNECTION_DEVICE_TYPE + "': " + device);
                }
                Device connectionDevice = connectionDevices[0];
                log.fine("Using first discovered WAN connection device: " + connectionDevice);

                Service ipConnectionService = connectionDevice.findService(new UDAServiceType("WANIPConnection", 1));
                Service pppConnectionService = connectionDevice.findService(new UDAServiceType("WANPPPConnection", 1));

                if (ipConnectionService == null && pppConnectionService == null) {
                    log.fine("IGD doesn't support IP or PPP WAN connection service: " + device);
                }
                routerService = ipConnectionService != null ? ipConnectionService : pppConnectionService;

                getExternalIP();

                try {
                    Log.i(TAG, "router: " + upnpService.get().getRouter().toString() + ", en=" + upnpService.get().getRouter().isEnabled() + ", conf=" + upnpService.get().getRouter().getConfiguration().toString());
                } catch (RouterException e) {
                    e.printStackTrace();
                }
            }

            String name = device.getDetails() != null && device.getDetails().getFriendlyName() != null
                            ? device.getDetails().getFriendlyName()
                            : device.getDisplayString();
            // Display a little star while the device is being loaded (see performance optimization earlier)
            name = device.isFullyHydrated() ? name : name + " *";
            Log.i(TAG, name);
        }

        public void deviceRemoved(final Device device) {
            Log.i(TAG, "DEVICE REMOVED");
            String name = device.getDetails() != null && device.getDetails().getFriendlyName() != null
                    ? device.getDetails().getFriendlyName()
                    : device.getDisplayString();
            // Display a little star while the device is being loaded (see performance optimization earlier)
            name = device.isFullyHydrated() ? name : name + " *";
            Log.i(TAG, name);
        }
    }

    //method for getting local ip
    private String getIpAddress() {
        String ip = "";
        try {
            Enumeration<NetworkInterface> enumNetworkInterfaces = NetworkInterface
                    .getNetworkInterfaces();
            while (enumNetworkInterfaces.hasMoreElements()) {
                NetworkInterface networkInterface = enumNetworkInterfaces
                        .nextElement();
                Enumeration<InetAddress> enumInetAddress = networkInterface
                        .getInetAddresses();
                while (enumInetAddress.hasMoreElements()) {
                    InetAddress inetAddress = enumInetAddress.nextElement();

                    if (inetAddress.isSiteLocalAddress()) {
                        ip +=inetAddress.getHostAddress();
                    }
                }
            }
        } catch (SocketException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            ip += "Something Wrong! " + e.toString() + "\n";
        }
        return ip;
    }

    public void getExternalIP(){
        upnpService.getControlPoint().execute(
                //new GetExternalIP(service) {
                new GetExternalIP(routerService) {

                    @Override
                    protected void success(String externalIPAddress) {
                        Log.i(TAG, "External IP is: " + externalIPAddress);
                        final String str = externalIPAddress;
                        runOnUiThread(new Runnable() {
                            @Override
                            public void run() {
                                txtRemote.setText("Public: " + str + ":" + port);
                            }
                        });

                    }

                    @Override
                    public void failure(ActionInvocation invocation,
                                        UpnpResponse operation,
                                        String defaultMsg) {
                        Log.i(TAG, "Cannot get external IP is");
                    }
                }
        );
    }

}
// DOC:CLASS_END
